import "RobYap.CleanChatWindows.Windows";

-- Requires Color Picker: https://www.lotrointerface.com/downloads/info860-ClassAnotherColorPicker.html

-- Note that /tell messages do not include the sender's name, so you only see what they typed
-- As such, the Tell ChatType is not available/presented anywhere in this plugin

_G.CHATTYPE_EventBroadcast = 24.000000;
_G.CHATTYPE_NPC_Say        = 500.000000;
_G.CHATTYPE_Rolls          = 400.000000;

_G.WindowList    = {};
_G.DebugMode     = false;

local ignoreList = {};
local filterList = {};
local snoozeList = {};
local savedWindowSettings = {};

_G.ChannelNames = {};
_G.ChannelNames[Turbine.ChatType.Advancement]   = "Advancement";
_G.ChannelNames[Turbine.ChatType.Death]         = "Death";
_G.ChannelNames[Turbine.ChatType.Emote]         = "Emote";
_G.ChannelNames[Turbine.ChatType.EnemyCombat]   = "Enemy Combat";
_G.ChannelNames[Turbine.ChatType.Error]         = "Error";
_G.ChannelNames[_G.CHATTYPE_EventBroadcast]     = "Event Broadcast";
_G.ChannelNames[Turbine.ChatType.Fellowship]    = "Fellowship";
_G.ChannelNames[Turbine.ChatType.Kinship]       = "Kinship";
_G.ChannelNames[Turbine.ChatType.LFF]           = "LFF";
_G.ChannelNames[Turbine.ChatType.FellowLoot]    = "Loot (Fellow)";
_G.ChannelNames[Turbine.ChatType.SelfLoot]      = "Loot (Self)";
_G.ChannelNames[Turbine.ChatType.OOC]           = "OOC";
_G.ChannelNames[Turbine.ChatType.Officer]       = "Officer";
_G.ChannelNames[Turbine.ChatType.PlayerCombat]  = "Player Combat";
_G.ChannelNames[Turbine.ChatType.Quest]         = "Quest";
_G.ChannelNames[Turbine.ChatType.Raid]          = "Raid";
_G.ChannelNames[Turbine.ChatType.Regional]      = "Regional";
_G.ChannelNames[Turbine.ChatType.Roleplay]      = "Roleplay";
_G.ChannelNames[_G.CHATTYPE_Rolls]              = "Rolls Only"
_G.ChannelNames[Turbine.ChatType.Say]           = "Say (Player)";
_G.ChannelNames[_G.CHATTYPE_NPC_Say]            = "Say (Non-Player)";
_G.ChannelNames[Turbine.ChatType.Standard]      = "Standard";
_G.ChannelNames[Turbine.ChatType.Tell]          = "Tell  *API Limitation, Cannot Show Name";
_G.ChannelNames[Turbine.ChatType.Trade]         = "Trade";
_G.ChannelNames[Turbine.ChatType.Unfiltered]    = "Unfiltered";
_G.ChannelNames[Turbine.ChatType.UserChat1]     = "UserChat1";
_G.ChannelNames[Turbine.ChatType.UserChat2]     = "UserChat2";
_G.ChannelNames[Turbine.ChatType.UserChat3]     = "UserChat3";
_G.ChannelNames[Turbine.ChatType.UserChat4]     = "UserChat4";
_G.ChannelNames[Turbine.ChatType.UserChat5]     = "UserChat5";
_G.ChannelNames[Turbine.ChatType.UserChat6]     = "UserChat6";
_G.ChannelNames[Turbine.ChatType.UserChat7]     = "UserChat7";
_G.ChannelNames[Turbine.ChatType.UserChat8]     = "UserChat8";
_G.ChannelNames[Turbine.ChatType.World]         = "World";

local savedSettings = Turbine.PluginData.Load(Turbine.DataScope.Character, "CleanChatWindows", function(f,a) end);

if savedSettings ~= nil then
  ignoreList          = savedSettings["IgnoreList"];
  filterList          = savedSettings["FilterList"];
  savedWindowSettings = savedSettings["WindowSettings"];
end

function DebugWrite(msg)
  if _G.DebugMode then
    Turbine.Shell.WriteLine(msg);
  end
end

---@diagnostic disable-next-line: lowercase-global
function iif ( cond , T , F )
  if cond then return T else return F end
end

---@diagnostic disable-next-line: lowercase-global
function split (inputstr, sep)
  if sep == nil then
          sep = "%s"
  end
  local t={}
  for str in string.gmatch(inputstr, "([^"..sep.."]+)") do
          table.insert(t, str)
  end
  return t
end

---@diagnostic disable-next-line: lowercase-global
function spairs(t, order)
  -- collect the keys
  local keys = {};
  for k in pairs(t) do keys[#keys+1] = k end;

  -- if order function given, sort by it by passing the table and keys a, b,
  -- otherwise just sort the keys 
  if order then
      table.sort(keys, function(a,b) return order(t, a, b) end);
  else
      table.sort(keys);
  end

  -- return the iterator function
  local i = 0;
  return function()
      i = i + 1;
      if keys[i] then
          return keys[i], t[keys[i]];
      end
  end
end

local next = next;

-- If no saved windows, make the default
if next(savedWindowSettings) == nil then
  -- Create new chat window and add to list
  table.insert(
    _G.WindowList,
    Windows.CleanChatWindow({ Name = "Player Say"})
  );
else
  for key,value in pairs(savedWindowSettings) do
      -- Create chat window and add to list
      table.insert(
        _G.WindowList,
        Windows.CleanChatWindow(savedWindowSettings[key])
      )
    --end
  end
end

-- Forces correct initial state of scrollbar for each window
for key,value in pairs(_G.WindowList) do
  _G.WindowList[key]:SetText(" ");
  _G.WindowList[key]:SetText("");
end

-- Create watcher to react to certain global key events
KeyWatcher = Turbine.UI.Control();
KeyWatcher.HudVisible = true;
KeyWatcher:SetWantsKeyEvents(true);

---@diagnostic disable-next-line: duplicate-set-field
KeyWatcher.KeyDown = function(sender, args)

  if (args.Action == Turbine.UI.Lotro.Action.UI_Toggle) then
    KeyWatcher.HudVisible = not(KeyWatcher.HudVisible)

    -- Call each window to process to toggle visibility
    for key,value in pairs(_G.WindowList) do
      if _G.WindowList[key]._settings.Active then
        _G.WindowList[key]:SetVisible(KeyWatcher.HudVisible);
      end
    end
  end

end

-- Create function to handle chat message events
ChatReceived = function(sender, args)

  --for innerKey,innerValue in pairs(args) do
  --  Turbine.Engine.ScriptLog(innerKey .. ": " .. args[innerKey]);
  --end
  --Turbine.Engine.ScriptLog(args.Message);

  local ignore = false;
  local next = next;

  -- If ignored characters exist
  if next(ignoreList) ~= nil then
    local msg = string.lower(args.Message:gsub("%b<>", ""));
    -- Check ignoreList, set flag if current message from ignored character
    for key,value in pairs(ignoreList) do
      if string.find(msg, ignoreList[key] .. ": '") or string.starts(msg, ignoreList[key].." ") then
        ignore = true;
        break;
      end
    end
  end

  if not ignore and next(filterList) ~= nil then
    local msg = string.lower(args.Message:gsub("%b<>", ""));
    -- Check filterList, set flag if current message has filtered keywords
    for key,value in pairs(filterList) do
      if string.find(msg, string.lower(filterList[key])) then
        ignore = true;
        break;
      end
    end
  end

  if not ignore and next(snoozeList) ~= nil and #snoozeList > 0 then
    -- Check filterList, set flag if current message has filtered keywords
    local msg = string.lower(args.Message:gsub("%b<>", ""));
    for i=#snoozeList,1,-1 do
      if string.find(msg, string.lower(snoozeList[i].Name) .. ": '") or string.starts(msg, string.lower(snoozeList[i].Name).." ") then
        if snoozeList[i].SnoozeUntil > Turbine.Engine.GetLocalTime() then
          ignore = true;
        else
          table.remove(snoozeList, i);
        end
        break;
      end
    end
  end

  if not ignore then
    for innerKey,innerValue in pairs(_G.WindowList) do
      -- Call each window to process the chat message
      _G.WindowList[innerKey]:ProcessChatMessage(sender, args);
    end
  end

  --Turbine.Engine.ScriptLog(msg);
end


function AddCallback(object, event, callback)

  if (object[event] == nil) then
    object[event] = callback;
  else
    if (type(object[event]) == "table") then
      table.insert(object[event], callback);
    else
      object[event] = {object[event], callback};
    end
  end
  return callback;

end


function RemoveCallback(object, event, callback)

  if (object[event] == callback) then
    object[event] = nil;
  else
    if (type(object[event]) == "table") then
      ---@diagnostic disable-next-line: deprecated
      local size = table.getn(object[event]);
      for i = 1, size do
        if (object[event][i] == callback) then
          table.remove(object[event], i);
          break;
        end
      end
    end
  end

end

function Save()
  local windowSettings = {};
  for key,value in pairs(_G.WindowList) do
    table.insert(windowSettings, _G.WindowList[key]._settings);
  end

  Turbine.PluginData.Save( Turbine.DataScope.Character, "CleanChatWindows", {IgnoreList = ignoreList, FilterList = filterList, WindowSettings = windowSettings}, function(f,a) end )

  Turbine.Shell.WriteLine( "<rgb=#777777>Clean Chat Windows Data Saved</rgb>" )
end

-- Subscribe to chat message event
AddCallback(Turbine.Chat, "Received", ChatReceived)

-- Unload, unsubscribe from chat message event
---@diagnostic disable-next-line: duplicate-set-field
Plugins["CleanChatWindows"].Unload = function( sender, args )
  RemoveCallback(Turbine.Chat, "Received", ChatReceived);
  Save();
end

function string.starts(String,Start)
  return string.sub(String,1,string.len(Start))==Start
end

--Register a command line...
---@diagnostic disable-next-line: lowercase-global
shellCommand = Turbine.ShellCommand();

---@diagnostic disable-next-line: duplicate-set-field
function shellCommand:Execute(command, args)

  if string.sub(args,1,string.len("new ")) == "new " then

    local exists = false;
    local name = args:gsub("new ", "");
    name = (name:gsub("^%s*(.-)%s*$", "%1"));

    if name == nil or name == "" then
      Turbine.Shell.WriteLine("Window name required")
    else
      for key,value in pairs(_G.WindowList) do
        if _G.WindowList[key]._settings.Name == name then
          exists = true;
          Turbine.Shell.WriteLine("Window '" .. name .. "' already exists")
        end
      end

      if not exists then
        -- Create new chat window and add to list
        table.insert(
          _G.WindowList,
          Windows.CleanChatWindow({ Name = name })
        );
        Save();
      end
    end

  elseif string.sub(args,1,string.len("toggle ")) == "toggle " then

    local exists = false;
    local name = args:gsub("toggle ", "");
    name = string.lower((name:gsub("^%s*(.-)%s*$", "%1")));

    if name == nil or name == "" then
      Turbine.Shell.WriteLine("Name required")
    else
      for key,value in pairs(_G.WindowList) do
        if string.lower(_G.WindowList[key]._settings.Name) == name then
          exists = true;
          if _G.WindowList[key]._settings.Active then
            _G.WindowList[key]._settings.Active = false;
            Save();
            _G.WindowList[key]:Close();
          else
            _G.WindowList[key]._settings.Active = true;
            local settings = _G.WindowList[key]._settings;
            _G.WindowList[key] = Windows.CleanChatWindow(settings);
          end
        end
      end

      if not exists then
        Turbine.Shell.WriteLine("Window '" .. name .. "' does not exist")
      end
    end

  elseif args == "list" then

    Turbine.Shell.WriteLine("<rgb=#ed7342>Chat Windows</rgb>")
    Turbine.Shell.WriteLine("<rgb=#ed7342>----------------------------</rgb>")
    local windowNames = {}
    for key,value in pairs(_G.WindowList) do
      local active = " (Active)";
      if not value._settings.Active then
        active = " (Disabled)";
      end
      table.insert(windowNames, value._settings.Name..active);
    end
    table.sort(windowNames)
    for i,windowName in ipairs(windowNames) do
      Turbine.Shell.WriteLine(windowName);
    end

	elseif string.sub(args,1,string.len("ignore ")) == "ignore " then

    local name = args:gsub("ignore ", "");
    local exists = false;
    if next(ignoreList) ~= nil then
      for key,value in pairs(ignoreList) do
        if string.lower(ignoreList[key]) == string.lower(name) then
          exists = true;
          break;
        end
      end
    end

    if exists then
      Turbine.Shell.WriteLine(name .. " already exists in ignore list");
    else
      table.insert(ignoreList, name);
      Turbine.Shell.WriteLine(name .. " added to ignore list");
      Save();
    end

	elseif args == "ignorelist" then

    Turbine.Shell.WriteLine("<rgb=#ed7342>Ignored Characters</rgb>")
    Turbine.Shell.WriteLine("<rgb=#ed7342>----------------------------</rgb>")
    local ignoredNames = {}
    for key,value in pairs(ignoreList) do
      table.insert(ignoredNames, value);
    end
    table.sort(ignoredNames)
    for i,ignoredName in ipairs(ignoredNames) do
      Turbine.Shell.WriteLine(ignoredName);
    end

  elseif string.sub(args,1,string.len("allow ")) == "allow " then

    local exists = false;
    local name = args:gsub("allow ", "");
    for i=#ignoreList,1,-1 do
      if string.lower(ignoreList[i]) == string.lower(name) then
          table.remove(ignoreList, i);
          exists = true;
          Turbine.Shell.WriteLine(name .. " removed from ignore list");
          Save();
      end
    end

    if not exists then
      Turbine.Shell.WriteLine(name .. " did not exist in ignore list")
    end


  elseif string.sub(args,1,string.len("filter ")) == "filter " then

    local text = args:gsub("filter ", "");
    local exists = false;
    if next(filterList) ~= nil then
      for key,value in pairs(filterList) do
        if string.lower(filterList[key]) == string.lower(text) then
          exists = true;
          break;
        end
      end
    end

    if exists then
      Turbine.Shell.WriteLine("'" .. text .. "' already exists in filter list");
    else
      table.insert(filterList, text);
      Turbine.Shell.WriteLine("'" .. text .. "' added to filter list");
      Save();
    end

	elseif args == "filterlist" then

    Turbine.Shell.WriteLine("<rgb=#ed7342>Filtered Text</rgb>")
    Turbine.Shell.WriteLine("<rgb=#ed7342>----------------------------</rgb>")
    local filteredStrings = {}
    for key,value in pairs(filterList) do
      table.insert(filteredStrings, value);
    end
    table.sort(filteredStrings)
    for i,filteredString in ipairs(filteredStrings) do
      Turbine.Shell.WriteLine(i .. ": " .. filteredString);
    end

  elseif string.sub(args,1,string.len("unfilter ")) == "unfilter " then

    local exists = false;
    local index = tonumber(args:gsub("unfilter ", ""));
    for i=#filterList,1,-1 do
      if i == index then
          table.remove(filterList, i);
          exists = true;
          Turbine.Shell.WriteLine("Entry " .. index .. " removed from filter list");
          Save();
      end
    end

    if not exists then
      Turbine.Shell.WriteLine("Entry " .. index .. " did not exist in filter list")
    end

	elseif string.sub(args,1,string.len("snooze ")) == "snooze " then

    local text = args:gsub("snooze ", "")
    if string.len(text) == 0 then
      Turbine.Shell.WriteLine("Missing name and duration");
      return;
    end
    local vals = split(text, " ");
    if next(vals) == nil then
      Turbine.Shell.WriteLine("Missing name and duration");
      return;
    end

    local name = "";
    local duration;
    local ct = 1;
    for key,value in pairs(vals) do
      if ct == 1 then
        name = string.lower(vals[key]);
      elseif ct == 2 then
        duration = tonumber(vals[key]);
      end
      ct = ct + 1;
    end
    if string.len(name) == 0 then
      Turbine.Shell.WriteLine("Missing name");
    end
    if duration == nil or duration < 0 then
      duration = 60;
    end

    local exists = false;
    if next(snoozeList) ~= nil then
      for key,value in pairs(snoozeList) do
        if string.lower(snoozeList[key]) == string.lower(name) then
          exists = true;
          break;
        end
      end
    end

    if exists then
      Turbine.Shell.WriteLine("'" .. name .. "' already exists in snooze list");
    else
      table.insert(snoozeList, { Name = string.lower(name), SnoozeUntil = Turbine.Engine.GetLocalTime() + ( duration * 60 )});
      Turbine.Shell.WriteLine("'" .. name .. "' added to snooze list");
    end

  elseif args == "clearsnooze" then

    for i=#snoozeList,1,-1 do
      table.remove(snoozeList, i);
    end

    Turbine.Shell.WriteLine("All snoozed characters cleared.")

  elseif args == "toggledebug" then

    _G.DebugMode = not _G.DebugMode;

    if _G.DebugMode then
      Turbine.Shell.WriteLine("Debug mode enabled.")
    else
      Turbine.Shell.WriteLine("Debug mode disabled.")
    end

  elseif args == "counts" then

    for key,value in pairs(_G.WindowList) do
      if _G.WindowList[key]._settings.Active then
        Turbine.Shell.WriteLine("<rgb=#ed7342>Counts for window: " .. _G.WindowList[key]._settings.Name .. "</rgb>")
        Turbine.Shell.WriteLine("<rgb=#ed7342>---------------------------------------------</rgb>")
        Turbine.Shell.WriteLine(_G.WindowList[key]:GetEntityMessageCounts(20));
      end
    end

	else
		Turbine.Shell.WriteLine("<rgb=#ffffff>Clean Chat Windows Commands</rgb>")
    Turbine.Shell.WriteLine("<rgb=#cccccc>/ccw new {window name} - creates a new chat window</rgb>")
    Turbine.Shell.WriteLine("<rgb=#cccccc>/ccw toggle {window name} - enables/disables chat window</rgb>")
    Turbine.Shell.WriteLine("<rgb=#cccccc>/ccw list - lists all chat windows</rgb>")
		Turbine.Shell.WriteLine("<rgb=#cccccc>/ccw ignore {character} - filters a character from all windows</rgb>")
    Turbine.Shell.WriteLine("<rgb=#cccccc>/ccw ignorelist - lists ignored characters</rgb>")
    Turbine.Shell.WriteLine("<rgb=#cccccc>/ccw snooze {character} {minutes} - ignores a character from all windows for X minutes (session only)</rgb>")
    Turbine.Shell.WriteLine("<rgb=#cccccc>/ccw clearsnooze  - removes all snooozed characters</rgb>")
    Turbine.Shell.WriteLine("<rgb=#cccccc>/ccw allow {character} - removes a previously ignored character</rgb>")
    Turbine.Shell.WriteLine("<rgb=#cccccc>/ccw filter {text} - filters messages from all windows containing the text</rgb>")
    Turbine.Shell.WriteLine("<rgb=#cccccc>/ccw filterlist - lists filtered text</rgb>")
    Turbine.Shell.WriteLine("<rgb=#cccccc>/ccw unfilter {number} - removes a text filter (get number from /filterlist)</rgb>")
    Turbine.Shell.WriteLine("<rgb=#cccccc>/ccw counts - lists top 20 message counts by entity name, for all windows</rgb>")
	end

end

Turbine.Shell.AddCommand('ccw', shellCommand)